//------------------------------------------------------------------------------
//	$Id: shrink-to-fit.js,v 1.24 2022/10/22 04:07:54 wolf Exp $
//------------------------------------------------------------------------------
//	Copyright (c) 2022 Wolfgang Stanglmeier <software@netzwolf.info>
//	This program is free software. It comes without any warranty.
//	You can redistribute it and/or modify it under the terms of the
//	"Do What The Fuck You Want To Public License" (WTFPL) Version 2,
//	as published by Sam Hocevar. See http://www.wtfpl.net/
//------------------------------------------------------------------------------
//	change font-size on marked elements to fit element into parent element
//------------------------------------------------------------------------------
//	load script using "defer" or at end of body
//------------------------------------------------------------------------------

/*jsl:option explicit*/

'use strict';

(function( ){

	//----------------------------------------------------------------------
	//	base config
	//----------------------------------------------------------------------

	const MinFontSize = 1;
	const MaxFontSize = 100;
	const FontSizeRes = 0.01;

	//----------------------------------------------------------------------
	//	element containing configuration
	//----------------------------------------------------------------------

	const configElement = document.getElementById( 'shrink-to-fit' );

	window.console.info( 'configElement: ' + configElement );

	//----------------------------------------------------------------------
	//	config
	//----------------------------------------------------------------------

	const ClassNames = (configElement &&
		configElement.getAttribute( 'data-shrink-to-fit-classes' ) ||
		'shrink-to-fit' ).split( /[,\s]+/ );

	window.console.info( 'ClassNames: ' + ClassNames );

	const Selector = configElement &&
		configElement.getAttribute( 'data-shrink-to-fit-selector' ) ||
		null;

	window.console.info( 'Selector: ' + Selector );

	const Delayed = (configElement &&
		configElement.getAttribute( 'data-shrink-to-fit-delayed' ) ||
		'display-math' ).split( /[,\s]+/ );

	window.console.info( 'Delayed: ' + Delayed );

	//----------------------------------------------------------------------
	//	create element lists
	//----------------------------------------------------------------------

	const lists = [];
	const types = [];

	//----------------------------------------------------------------------
	//	get elementslists from class name
	//----------------------------------------------------------------------

	for( var c=0; c < ClassNames.length; ++c ){

		const className = ClassNames[ c ];

		types.push( className );
		lists.push( document.getElementsByClassName( className ));
	}

	//----------------------------------------------------------------------
	//	get elementslists from selector
	//----------------------------------------------------------------------

	if( Selector ){

		try{
			lists.push( document.querySelectorAll( Selector ));
			types.push( Selector );

		} catch( exp ){

			window.console.error( 'shrink-to-fit.querySelectorAll: ' + exp );
		}
	}

	//----------------------------------------------------------------------
	//	nodes to be shrinked on demand
	//----------------------------------------------------------------------

	const objects   = [];
	const knownEles = [];

	for( var l=0; l < lists.length; ++l ){

		const elements = lists[ l ];

		window.console.info( types[ l ] + ' \u2192 ' + elements.length + ' elements.' );

		for( var e=0; e < elements.length; ++e ){

			const element = elements[e];

			//------------------------------------------------------
			//	accept each element only once
			//------------------------------------------------------

			if( knownEles.indexOf( element ) >= 0 ){ continue; }
			knownEles.push( element );

			//------------------------------------------------------
			//	create entry
			//------------------------------------------------------

			objects.push({
				type: types[l],
				elem: element,
				tout: null,
				iter: 0
			});
		}
	}

	//----------------------------------------------------------------------
	//	postprocess objects
	//----------------------------------------------------------------------

	if( !objects.length ){ return; }

	for( o in objects ){

		const object = objects[o];

		object.elem.style.fontSize = '100%';
		object.elem.setAttribute( 'data-display-style',
			object.elem.style.display || 'table' );
	}

	//----------------------------------------------------------------------
	//	update fontsize
	//----------------------------------------------------------------------

	const recalculateObject = function( object ){

		object.tout = null;

		//--------------------------------------------------------------
		//	available width of parent element (includes padding)
		//	effective width of element        (includes padding)
		//--------------------------------------------------------------

		var availableWidth = object.elem.parentElement.clientWidth;
		var requiredWidth  = object.elem.scrollWidth;

		//--------------------------------------------------------------
		//	handle margins, borders, padding
		//--------------------------------------------------------------

		if( 'getComputedStyle' in window ){

const pcs = window.getComputedStyle( object.elem.parentElement );
const ecs = window.getComputedStyle( object.elem );

availableWidth -= parseFloat( pcs.getPropertyValue( 'padding-left'      )) || 0;
availableWidth -= parseFloat( ecs.getPropertyValue( 'margin-left'       )) || 0;
availableWidth -= parseFloat( ecs.getPropertyValue( 'border-left-width' )) || 0;
availableWidth -= parseFloat( ecs.getPropertyValue( 'padding-left'      )) || 0;
availableWidth -= parseFloat( ecs.getPropertyValue( 'padding-right'     )) || 0;
availableWidth -= parseFloat( ecs.getPropertyValue( 'border-right-width')) || 0;
availableWidth -= parseFloat( ecs.getPropertyValue( 'margin-right'      )) || 0;
availableWidth -= parseFloat( pcs.getPropertyValue( 'padding-right'     )) || 0;

requiredWidth  -= parseFloat( ecs.getPropertyValue( 'padding-left'  )) || 0;
requiredWidth  -= parseFloat( ecs.getPropertyValue( 'padding-right' )) || 0;
		}

		//--------------------------------------------------------------
		//	restore display (set to 'block' at last iteration)
		//--------------------------------------------------------------

		object.elem.style.display =
			object.elem.getAttribute( 'data-display-style' );

		//--------------------------------------------------------------
		//	scale to exactly fit object.elem in parent.
		//	expand only at first iteration
		//--------------------------------------------------------------

		const scale = availableWidth / requiredWidth;

		//--------------------------------------------------------------
		//	expand during first iteration only
		//--------------------------------------------------------------

		if( scale >= 1 && object.iter > 1 ){ return; }

		//--------------------------------------------------------------
		//	compute new font size, done if unchanged
		//--------------------------------------------------------------

		const oldFontSize =
			parseFloat( object.elem.style.fontSize, 10 );
		const newFontSize = Math.max( MinFontSize, Math.min(
			MaxFontSize, FontSizeRes *
			Math.round( scale * oldFontSize / FontSizeRes )));

		if( oldFontSize === newFontSize ){ return; }

		//--------------------------------------------------------------
		//	set new font size
		//--------------------------------------------------------------

		object.elem.style.fontSize = newFontSize + '%';

		//--------------------------------------------------------------
		//	scale images too
		//--------------------------------------------------------------

		if( object.elem.getAttribute( 'data-shrink' ) == 'img' ){

			const images = object.elem.getElementsByTagName( 'img' );

			for( var i=0; i<images.length; ++i ){

				const img = images[i];

	const imgWidth  = parseInt( img.getAttribute( 'width' ),  10 );
	const imgHeight = parseInt( img.getAttribute( 'height' ), 10 );

				if( !imgWidth || !imgHeight ){ continue; }

	img.style.width  = Math.round( imgWidth  * newFontSize / 100 ) + 'px';
	img.style.height = Math.round( imgHeight * newFontSize / 100 ) + 'px';
			}
		}

		//--------------------------------------------------------------
		//	terminate after 4th iteration
		//--------------------------------------------------------------

		if( object.iter >= 4 ){ return; }

		//--------------------------------------------------------------
		//	request iteration
		//--------------------------------------------------------------

		object.elem.style.display = 'block';
		object.iter += 1;

		object.tout = window.setTimeout( function( ){
			recalculateObject( object );
		}, 1 );
	};

	//----------------------------------------------------------------------
	//	start updating a single object
	//----------------------------------------------------------------------

	function updateObject( object ){

		//--------------------------------------------------------------
		//	cancel running recalculations
		//--------------------------------------------------------------

		if( object.tout ){

			window.clearTimeout( object.tout );
			object.tout = null;
		}

		object.iter = 1;
		object.elem.style.fontSize = '100%';
		recalculateObject( object );
	}

	//----------------------------------------------------------------------
	//	update all objects
	//----------------------------------------------------------------------

	function updateAllObjects( on_init ){

		for( var o in objects ){

			const object = objects[o];

			if( on_init === true &&
					Delayed.indexOf( object.type ) >= 0 ){
				continue;
			}

			updateObject( object );
		}
	}

	//----------------------------------------------------------------------
	//	on window width change update all objects
	//----------------------------------------------------------------------

	var windowWidth  = window.innerWidth;

	window.addEventListener( 'resize', function(){

		//--------------------------------------------------------------
		//	but 1st check if false event by a braindead mobile browser
		//--------------------------------------------------------------

		if( window.innerWidth == windowWidth ){ return; }
		windowWidth  = window.innerWidth;

		updateAllObjects( false );

	}, false );

	//----------------------------------------------------------------------
	//	update object on content change
	//----------------------------------------------------------------------

	function attachObserver( object ){

		const config = {
			'characterData': true,
			'childList': true,
			'subtree': true
		};

		const observer = new window.MutationObserver( function(){

			if( object.obsv ){ window.clearTimeout( object.obsv ); }

			const delay =
				Delayed.indexOf( object.type ) >= 0 ? 1000 : 30;

			object.obsv = window.setTimeout( function(){

				updateObject( object );
			}, delay );
		});

		observer.observe( object.elem, config );
	}

	if( 'MutationObserver' in window ){

		for( var o in objects ){

			attachObserver( objects[o] );
		}
	}

	//----------------------------------------------------------------------
	//	initial update
	//----------------------------------------------------------------------

	updateAllObjects( true );
})();

//------------------------------------------------------------------------------
//	$Id: shrink-to-fit.js,v 1.24 2022/10/22 04:07:54 wolf Exp $
//------------------------------------------------------------------------------
